;;;; resolver.lisp
;;;; By Dave Roberts
;;;; Copyright (C) 2004, Dave Roberts, All Rights Reserved
;;;;
;;;; This software is distributed under the Lesser Gnu Public License (LGPL).
;;;; See the file "COPYING" that came with this package.

(in-package #:resolver)
(declaim (optimize (speed 3) (safety 0)))

;; Define the Lisp version of the C function
(uffi:def-function ("res_query" res-query)
    ((dname (:cstring))
     (class :int)
     (type :int)
     (answer (* :unsigned-byte))
     (anslen :int))
  :returning :int)

;; Types for various DECLARE/DECLAIM optimizations
(uffi:def-array-pointer uffi-unsigned-byte-pointer :unsigned-byte)
(uffi:def-type unsigned-byte-pointer '(* :unsigned-byte))
(deftype buffer-index () `(integer 0 ,(1- (expt 2 14))))
(deftype uint8 () '(unsigned-byte 8))
(deftype uint16 () '(unsigned-byte 16))
(deftype uint32 () '(unsigned-byte 32))
(deftype dns-pointer () `(integer 0  ,(1- (expt 2 14))))

;; have a condition to signal our errors.
(define-condition decoding-error (error)
  ((position :initarg :pos :reader pos)))

(define-condition out-of-bounds-error (decoding-error)
  ((index :initarg :index :reader index)))

;; get-byte is *the* decoding primitive, so make sure it's inlined
;; everywhere possible for maximum speed
(declaim (inline get-byte)
	 (ftype (function (unsigned-byte-pointer buffer-index buffer-index) uint8) 
		get-byte))

(defun get-byte (buffer size index)
  "Get a byte from the foreign object array at the specified index."
  (declare (optimize (speed 3) (safety 0) (space 0))
 	   (unsigned-byte-pointer buffer)
 	   (buffer-index size)
 	   (buffer-index index))
  (unless (< -1 index size)
    (error 'out-of-bounds-error :index index))
  (the uint8 (uffi:deref-array buffer 'uffi-unsigned-byte-pointer index)))

(declaim (ftype (function (unsigned-byte-pointer buffer-index buffer-index) uint16) get16))

(defun get16 (buffer size index)
  "Get a 16-bit quantity from the foreign object array, starting at
the specified index, in big-endian byte order."
  (declare (optimize (speed 3) (safety 0) (space 0))
	   (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (logior (ash (get-byte buffer size index) 8)
	  (get-byte buffer size (1+ index))))

(declaim (ftype (function (unsigned-byte-pointer buffer-index buffer-index) uint32) get32))

(defun get32 (buffer size index)
  "Get a 32-bit quantity from the foreign object array, starting at
the specified index, in big-endian byte order."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (the uint32 (logior (ash (get16 buffer size index) 16)
		      (get16 buffer size (+ index 2)))))

(defun decode-pointer (buffer size index)
  "Decode a DNS label pointer, starting at the specified offset,
returning the pointer value."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (logior (ash (logand (get-byte buffer size index) #X3F) 8)
	  (get-byte buffer size (1+ index))))

(defun decode-label (buffer size index)
  "Decode one label of a larger DNS name, starting at the specified
index within the buffer. Returns the label and the index of the byte
following the label."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (let* ((length (get-byte buffer size index))
	 (label (make-string length))
	 (start (1+ index)))
    (dotimes (i length)
      (setf (char label i)
	    (code-char (get-byte buffer size (+ start i)))))
    (values label (+ index length 1))))

(defun decode-name (buffer size index)
  "Decode the DNS name, starting at the specified index. Correctly
handles pointer traversal. Returns the name and the index of the byte
following the name."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (do* ((i index)
	(length (get-byte buffer size i) (get-byte buffer size i))
	(name ""))
       ((zerop length) (values name (1+ i)))
    (when (>= length #xC0)
      (let ((pointer (decode-pointer buffer size i)))
	(if (< pointer index)
	    (return (values (concatenate 'string name 
					 (decode-name buffer size pointer))
			    (+ i 2)))
	    (error 'decoding-error :pos i))))
    (multiple-value-bind (label new-index) (decode-label buffer size i)
      (setq name (concatenate 'string name label "."))
      (setq i new-index))))

(defun decode-ipv4-addr (buffer size index)
  "Decode an IPv4 address at the specified index. Returns the address
as a vector of four bytes and the index of the byte following the
address."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (let ((addr (make-array 4 :element-type 'fixnum)))
    (dotimes (i 4) (setf (aref addr i) (get-byte buffer size (+ index i))))
    addr))

(defun decode-header (buffer size)
  "Decode the header of the reply packet, returning a list of symbolic
constants representing the id, opcode, rcode, flags, and section
counts."
  (let ((id (get16 buffer size 0))
	(ops-flgs (get16 buffer size 2))
	(qdcount (decode-query-count buffer size))
	(ancount (decode-answer-count buffer size))
	(nscount (decode-name-server-count buffer size))
	(arcount (decode-additional-count buffer size)))
    (let ((qr (= (ldb (byte 1 15) ops-flgs) 1))
	  (opcode (opcode->symbol (ldb (byte 4 11) ops-flgs)))
	  (aa (= (ldb (byte 1 10) ops-flgs) 1))
	  (tc (= (ldb (byte 1 9) ops-flgs) 1))
	  (rd (= (ldb (byte 1 8) ops-flgs) 1))
	  (ra (= (ldb (byte 1 7) ops-flgs) 1))
	  (rcode (rcode->symbol (ldb (byte 4 0) ops-flgs))))
      (list id qr opcode aa tc rd ra rcode qdcount ancount nscount arcount))))

(defun decode-query-count (buffer size)
  "Given a buffer holding a DNS message, decode the query count and
return it."
  (get16 buffer size 4))

(defun decode-answer-count (buffer size)
  "Given a buffer holding a DNS message, decode the answer count and
return it."
  (get16 buffer size 6))

(defun decode-name-server-count (buffer size)
  "Given a buffer holding a DNS message, decode the name server count
and return it."
  (get16 buffer size 8))

(defun decode-additional-count (buffer size)
  "Given a buffer holding a DNS message, decode the name server count
and return it."
  (get16 buffer size 10))

(defun decode-query (buffer size index)
  "Decode a question starting at the specified index in the
buffer. Returns the question elements as a list and the index of the
byte following the question."
  (multiple-value-bind (qname idx-of-type) (decode-name buffer size index)
    (values (list qname 
		  (typecode->symbol (get16 buffer size idx-of-type)) 
		  (classcode->symbol (get16 buffer size (+ 2 idx-of-type))))
	    (+ 4 idx-of-type))))

(defun decode-queries (buffer size)
  "Decodes the entire question section of the message and returns a
list of question elements and the index of the byte following the last
question in the buffer."
  (let ((queries nil)
	(index 12)
	query)
    (dotimes (i (decode-query-count buffer size) (values (nreverse
							  queries) index))
      (multiple-value-setq (query index) (decode-query buffer
						       size 
						       index))
      (push query queries))))

(defun decode-generic-rdata (buffer size index)
  "Decode the RDATA portion of a resource record. Return the data as a
list of byte values and the index of the byte following the RDATA."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (let ((length (get16 buffer size index))
	(result nil))
    (dotimes (i length (values (list (nreverse result)) (+ 2 index length)))
      (push (get-byte buffer size (+ index 2 i)) result))))

(defun decode-domain-name-rdata (buffer size index)
  "Decodes a generic piece of RDATA that consists of a simple domain
name. Used for decoding such things as NS, CNAME, MD, MB, etc."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (let ((length (get16 buffer size index)))
    (multiple-value-bind (cname end-idx) (decode-name buffer size (+ 2 index))
      (unless (= (+ index length 2) end-idx)
	(error 'decoding-error :pos index))
      (values (list cname) end-idx))))

(defun decode-text-rdata (buffer size index)
  "Decodes a generic piece of RDATA that consists of a
<character-string>. Used for such things as TXT and HINFO RDATA."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (let* ((length (get16 buffer size index))
	 (text (make-string length)))
    (dotimes (i length) (setf (char text i)
			      (code-char (get-byte buffer size
						   (+ index 2 i)))))
    (values (list text) (+ index 2 length))))


(defun decode-a-rdata (buffer size index)
  "Decode the RDATA portion of a address (A) resource record. Returns
the address as a 4-element vector and the index of the byte
following the RDATA."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (let ((length (get16 buffer size index))
	(addr (decode-ipv4-addr buffer size (+ 2 index))))
    (unless (= length 4)
      (error 'decoding-error :pos index))
    (values (list addr) (+ 6 index))))

(defun decode-soa-rdata (buffer size index)
  "Decode SOA RDATA. Returns the data as a list and the index of the
byte following the RDATA."
  (let ((length (get16 buffer size index)))
    (multiple-value-bind (mname rname-start-idx) 
	(decode-name buffer size (+ index 2))
      (multiple-value-bind (rname serial-start-idx)
	  (decode-name buffer size rname-start-idx)
	(let ((serial (get32 buffer size serial-start-idx))
	      (refresh (get32 buffer size (+ serial-start-idx 4)))
	      (retry (get32 buffer size (+ serial-start-idx 8)))
	      (expire (get32 buffer size (+ serial-start-idx 12)))
	      (minimum (get32 buffer size (+ serial-start-idx 16)))
	      (end-idx (+ serial-start-idx 20)))
	  (unless (= (+ index 2 length) end-idx)
	    (error 'decoding-error :pos index))
	  (values (list mname rname serial refresh retry expire minimum)
		  end-idx))))))

;;(defun decode-wks-rdata (buffer size index)

(defun decode-minfo-rdata (buffer size index)
  "Decode two domain names as part of MINFO RDATA."
  (let ((length (get16 buffer size index)))
    (multiple-value-bind (rmailbx emailbx-start-idx)
	(decode-name buffer size (+ index 2))
      (multiple-value-bind (emailbx next-idx) 
	  (decode-name buffer size emailbx-start-idx)
	(unless (= (+ index 2 length) next-idx)
	  (error 'decoding-error :pos index))
	(values (list rmailbx emailbx) next-idx)))))

(defun decode-mx-rdata (buffer size index)
  "Decode MX RDATA. Returns the data values as a list, along with the
index of the byte following the RDATA."
  (declare (unsigned-byte-pointer buffer)
	   (buffer-index size)
	   (buffer-index index))
  (let ((length (get16 buffer size index))
	(preference (get16 buffer size (+ 2 index))))
    (multiple-value-bind (exchange end-idx) (decode-name buffer size
							 (+ 4 index))
      (unless (= (+ index 2 length) end-idx)
	(error 'decoding-error :pos index))
      (values (list preference exchange) end-idx))))

(defmacro define-mappers (fwd rev maplist &optional otherw1 otherw2)
  "A small macro to define two mapping functions between numbers and
symbols. Basically creates two functions with CASE forms to do the
mapping."
  (let ((farg (gensym))
	(rarg (gensym)))
    `(progn
      (defun ,fwd (,farg)
	(case ,farg
	  ,@maplist
	  (otherwise ,otherw1)))
      (defun ,rev (,rarg)
	(case ,rarg
	  ,@(mapcar #'reverse maplist)
	  (otherwise ,otherw2))))))

(define-mappers typecode->symbol symbol->typecode
  ((1 :A)
   (2 :NS)
   (3 :MD)
   (4 :MF)
   (5 :CNAME)
   (6 :SOA)
   (7 :MB)
   (8 :MG)
   (9 :MR)
   (10 :NULL)
   (11 :WKS)
   (12 :PTR)
   (13 :HINFO)
   (14 :MINFO)
   (15 :MX)
   (16 :TXT)
   (252 :AXFR)
   (253 :MAILB)
   (254 :MAILA)
   (255 :ALL)) :UNKNOWN NIL)

(define-mappers classcode->symbol symbol->classcode
  ((1 :IN) (2 :CS) (3 :CH) (4 :HS) (255 :ANY))
  :UNKNOWN nil)

(define-mappers opcode->symbol symbol->opcode 
  ((0 :query) (1 :iquery) (2 :status))
  :reserved nil)

(define-mappers rcode->symbol symbol->rcode 
  ((0 :no-error)
   (1 :format-error)
   (2 :server-failure)
   (3 :name-error)
   (4 :not-implemented)
   (5 :refused)) :reserved nil)

(defun decode-rr (buffer size index)
  "Decode a resource record, starting at the specified index in the
buffer. Return each of the elements of the resource record as a
list. Returns the list and the index of the byte following the
resource record."
  ;; Decode all the generic stuff first: name, type, class, ttl
  (multiple-value-bind (dname idx-of-type) (decode-name buffer size index)
    (let* ((type (get16 buffer size idx-of-type))
	   (class (get16 buffer size (+ 2 idx-of-type)))
	   (ttl (get32 buffer size (+ 4 idx-of-type)))
	   (rdata-func (case type
			 (1 #'decode-a-rdata)
			 (2 #'decode-domain-name-rdata)	; NS
			 (3 #'decode-domain-name-rdata)	; MD
			 (4 #'decode-domain-name-rdata)	; MF
			 (5 #'decode-domain-name-rdata)	; CNAME
			 (6 #'decode-soa-rdata)
			 (7 #'decode-domain-name-rdata)	; MB
			 (8 #'decode-domain-name-rdata)	; MG
			 (9 #'decode-domain-name-rdata)	; MR
			 (10 #'decode-generic-rdata) ; NULL
			 (11 #'decode-generic-rdata) ; WKS: #'decode-wks-rdata)
			 (12 #'decode-domain-name-rdata) ; PTR
			 (13 #'decode-text-rdata) ; HINFO
			 (14 #'decode-minfo-rdata)
			 (15 #'decode-mx-rdata)
			 (16 #'decode-text-rdata) ; TXT
			 (otherwise #'decode-generic-rdata))))
      (multiple-value-bind (rdata end-idx)
	  (funcall rdata-func buffer size (+ 8 idx-of-type))
	(values (append 
		 (list dname (typecode->symbol type) (classcode->symbol class) 
		       ttl) 
		 rdata)
		end-idx)))))

(defun decode-rrs (buffer size index count)
  "Decode <count> resource records from the buffer, starting at
<index>. Returns a list of resource records and the index of the byte
following the last record."
  (let ((result nil)
	(index index)
	rr)
    (dotimes (i count (values (nreverse result) index))
      (multiple-value-setq (rr index) (decode-rr buffer size index))
      (push rr result))))

(defun decode-reply-message (buffer size)
  "Decode a complete reply message. Returns a 5-element list
containing lists of each of the DNS message sections (header, query,
answer, name server, and additional."
  (let ((header (decode-header buffer size))
	(ancount (decode-answer-count buffer size))
	(nscount (decode-name-server-count buffer size))
	(arcount (decode-additional-count buffer size)))
    (multiple-value-bind (questions index) (decode-queries buffer size)
      (multiple-value-bind (answers index)
	  (decode-rrs buffer size index ancount)
	(multiple-value-bind (name-servers index) 
	    (decode-rrs buffer size index nscount)
	  (multiple-value-bind (additional index) 
	      (decode-rrs buffer size index arcount)
	    (unless (= index size)
	      (error 'decoding-error :pos 0))
	    (list header questions answers name-servers additional)))))))

(defun do-query (domain type buffer size)
  "Does a DNS lookup for a record of the given <type> associated with
<domain> name and puts the result into the specified <buffer> of the
indicated <size>. Uses DNS IN class for all lookups. Returns the DNS
reply decoded into a list containing 5 sublists representing the
header, query, answer, name server, and additional DNS message
sections."
  (let ((typecode (symbol->typecode type)))
    (if (null typecode)
	(error "Bad type: ~A" type)
	(uffi:with-cstring (cdomain domain)
	  (let ((length (res-query cdomain 1
				     typecode buffer size)))
	    (if (< length 0)
		nil
		(decode-reply-message buffer length)))))))

;; Turn this back on for the API into the library. Want to make sure
;; that all parameters are checked.
(declaim (optimize (safety 3)))

(defun lookup (domain type)
  "Looks up the specified <domain> name and DNS record <type>. Uses
DNS IN class for all lookups. Returns the DNS reply decoded into a
list containing 5 sublists representing the header, query, answer,
name server, and additional DNS message sections."
  (let* ((len (1- (expt 2 14)))
	 (buffer (uffi:allocate-foreign-object :unsigned-byte len)))
    (unwind-protect (do-query domain type buffer len)
      (uffi:free-foreign-object buffer))))

(declaim (inline get-id get-qr get-opcode get-aa get-tc get-rd get-ra get-rcode
		 get-qdcount get-ancount get-nscount get-arcount
		 get-queries get-answers get-name-servers get-additional))

(defun get-id (reply)
  (nth 0 (first reply)))

(defun get-qr (reply)
  (nth 1 (first reply)))

(defun get-opcode (reply)
  (nth 2 (first reply)))

(defun get-aa (reply)
  (nth 3 (first reply)))

(defun get-tc (reply)
  (nth 4 (first reply)))

(defun get-rd (reply)
  (nth 5 (first reply)))

(defun get-ra (reply)
  (nth 6 (first reply)))

(defun get-rcode (reply)
  (nth 7 (first reply)))

(defun get-qdcount (reply)
  (nth 8 (first reply)))

(defun get-ancount (reply)
  (nth 9 (first reply)))

(defun get-nscount (reply)
  (nth 10 (first reply)))

(defun get-arcount (reply)
  (nth 11 (first reply)))

(defun get-queries (reply)
  (second reply))

(defun get-answers (reply)
  (third reply))

(defun get-name-servers (reply)
  (fourth reply))

(defun get-additional (reply)
  (fifth reply))

(defun get-domain (rr)
  (first rr))

(defun get-type (rr)
  (second rr))

(defun get-class (rr)
  (third rr))

(defun get-ttl (rr)
  (fourth rr))

(defun get-rdata (rr)
  (cddddr rr))

;;;;----------------------------------------------------------------------
;;;; Test code
;;;;----------------------------------------------------------------------

(defun dump-array (array size start end)
  "Show the array of data."
  (do ((i start (1+ i)))
      ((= i end))
    (format t "(~A:~A) " i (get-byte array size i))))

(defun corrupt-buffer (buffer length)
  "XOR a random byte of the buffer with a random number from 0 -
255. In other words, flip a few bits."
  (let ((index (random length))
	(pattern (random 256)))
    (setf (uffi:deref-array buffer 'uffi-unsigned-byte-pointer index)
	  (logxor pattern (uffi:deref-array buffer 'uffi-unsigned-byte-pointer index)))))

(defun test-corruption (domain type)
  "Puts the decoding code through the wringer. Basically, given a
domain name and type, this function performs between 10 and 19 random
cycles of retreiving DNS information into a buffer, and then
repeatedly corrupting and decoding it between 50 and 99 times. Any
decoding-errors or out-of-bounds-errors are dealt with silently. Any
other type of error is allowed to propagate and constitues failure of
the test."
  (let* ((len (1- (expt 2 14)))
	 (buffer (uffi:allocate-foreign-object :unsigned-byte len))
	 (typecode (symbol->typecode type)))
    (unwind-protect 
	 (if (null typecode)
	     (error "Bad type: ~A" type)
	     (dotimes (i (+ 10 (random 10)))
	       (uffi:with-cstring (cdomain domain)
		 (let ((length (res-query cdomain 1 typecode buffer len)))
		   (if (< length 0)
		       (error "DNS error.")
		       (dotimes (j (+ 50 (random 50)))
			 (block nil
			   (format t "~A," j)
			   (corrupt-buffer buffer length)
			   (handler-bind 
			       (((or decoding-error out-of-bounds-error)
				 #'(lambda (c) (return c)))) ; exit the block
			     (decode-reply-message buffer length)))))
		   (format t "~%")))))
      (uffi:free-foreign-object  buffer))))
